@@ -3,7 +3,7 @@
  *
  * Copyright (C) 2005-2009 Johan Dahlin <johan@gnome.org>
  *
- *   pygi-argument.c: GArgument - PyObject conversion functions.
+ *   pygi-argument.c: GIArgument - PyObject conversion functions.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -28,7 +28,7 @@
 
 #include <datetime.h>
 #include <pygobject.h>
-
+#include <pyglib-python-compat.h>
 
 static void
 _pygi_g_type_tag_py_bounds (GITypeTag   type_tag,
@@ -37,29 +37,29 @@ _pygi_g_type_tag_py_bounds (GITypeTag   type_tag,
 {
     switch (type_tag) {
         case GI_TYPE_TAG_INT8:
-            *lower = PyInt_FromLong (-128);
-            *upper = PyInt_FromLong (127);
+            *lower = PYGLIB_PyLong_FromLong (-128);
+            *upper = PYGLIB_PyLong_FromLong (127);
             break;
         case GI_TYPE_TAG_UINT8:
-            *upper = PyInt_FromLong (255);
-            *lower = PyInt_FromLong (0);
+            *upper = PYGLIB_PyLong_FromLong (255);
+            *lower = PYGLIB_PyLong_FromLong (0);
             break;
         case GI_TYPE_TAG_INT16:
-            *lower = PyInt_FromLong (-32768);
-            *upper = PyInt_FromLong (32767);
+            *lower = PYGLIB_PyLong_FromLong (-32768);
+            *upper = PYGLIB_PyLong_FromLong (32767);
             break;
         case GI_TYPE_TAG_UINT16:
-            *upper = PyInt_FromLong (65535);
-            *lower = PyInt_FromLong (0);
+            *upper = PYGLIB_PyLong_FromLong (65535);
+            *lower = PYGLIB_PyLong_FromLong (0);
             break;
         case GI_TYPE_TAG_INT32:
-            *lower = PyInt_FromLong (G_MININT32);
-            *upper = PyInt_FromLong (G_MAXINT32);
+            *lower = PYGLIB_PyLong_FromLong (G_MININT32);
+            *upper = PYGLIB_PyLong_FromLong (G_MAXINT32);
             break;
         case GI_TYPE_TAG_UINT32:
             /* Note: On 32-bit archs, this number doesn't fit in a long. */
             *upper = PyLong_FromLongLong (G_MAXUINT32);
-            *lower = PyInt_FromLong (0);
+            *lower = PYGLIB_PyLong_FromLong (0);
             break;
         case GI_TYPE_TAG_INT64:
             /* Note: On 32-bit archs, these numbers don't fit in a long. */
@@ -68,7 +68,7 @@ _pygi_g_type_tag_py_bounds (GITypeTag   type_tag,
             break;
         case GI_TYPE_TAG_UINT64:
             *upper = PyLong_FromUnsignedLongLong (G_MAXUINT64);
-            *lower = PyInt_FromLong (0);
+            *lower = PYGLIB_PyLong_FromLong (0);
             break;
         case GI_TYPE_TAG_FLOAT:
             *upper = PyFloat_FromDouble (G_MAXFLOAT);
@@ -160,6 +160,97 @@ _pygi_g_registered_type_info_check_object (GIRegisteredTypeInfo *info,
     return retval;
 }
 
+gint
+_pygi_g_type_interface_check_object (GIBaseInfo *info,
+                                     PyObject   *object)
+{
+    gint retval = 1;
+    GIInfoType info_type;
+
+    info_type = g_base_info_get_type (info);
+    switch (info_type) {
+        case GI_INFO_TYPE_CALLBACK:
+            if (!PyCallable_Check (object)) {
+                PyErr_Format (PyExc_TypeError, "Must be callable, not %s",
+                              object->ob_type->tp_name);
+                retval = 0;
+            }
+            break;
+        case GI_INFO_TYPE_ENUM:
+            retval = 0;
+            if (PyNumber_Check (object)) {
+                PyObject *number = PYGLIB_PyNumber_Long (object);
+                if (number == NULL)
+                    PyErr_Clear();
+                else {
+                    glong value = PYGLIB_PyLong_AsLong (number);
+                    int i;
+                    for (i = 0; i < g_enum_info_get_n_values (info); i++) {
+                        GIValueInfo *value_info = g_enum_info_get_value (info, i);
+                        glong enum_value = g_value_info_get_value (value_info);
+                        g_base_info_unref (value_info);
+                        if (value == enum_value) {
+                            retval = 1;
+                            break;
+                        }
+                    }
+                }
+            }
+            if (retval < 1)
+                retval = _pygi_g_registered_type_info_check_object (
+                             (GIRegisteredTypeInfo *) info, TRUE, object);
+            break;
+        case GI_INFO_TYPE_FLAGS:
+            if (PyNumber_Check (object)) {
+                /* Accept 0 as a valid flag value */
+                PyObject *number = PYGLIB_PyNumber_Long (object);
+                if (number == NULL)
+                    PyErr_Clear();
+                else {
+                    long value = PYGLIB_PyLong_AsLong (number);
+                    if (value == 0)
+                        break;
+                    else if (value == -1)
+                        PyErr_Clear();
+                }
+            }
+            retval = _pygi_g_registered_type_info_check_object (
+                         (GIRegisteredTypeInfo *) info, TRUE, object);
+            break;
+        case GI_INFO_TYPE_STRUCT:
+        {
+            GType type;
+
+            /* Handle special cases. */
+            type = g_registered_type_info_get_g_type ( (GIRegisteredTypeInfo *) info);
+            if (g_type_is_a (type, G_TYPE_CLOSURE)) {
+                if (!PyCallable_Check (object)) {
+                    PyErr_Format (PyExc_TypeError, "Must be callable, not %s",
+                                  object->ob_type->tp_name);
+                    retval = 0;
+                }
+                break;
+            } else if (g_type_is_a (type, G_TYPE_VALUE)) {
+                /* we can't check g_values because we don't have 
+                 * enough context so just pass them through */
+                break;
+            }
+
+            /* Fallback. */
+        }
+        case GI_INFO_TYPE_BOXED:
+        case GI_INFO_TYPE_INTERFACE:
+        case GI_INFO_TYPE_OBJECT:
+        case GI_INFO_TYPE_UNION:
+            retval = _pygi_g_registered_type_info_check_object ( (GIRegisteredTypeInfo *) info, TRUE, object);
+            break;
+        default:
+            g_assert_not_reached();
+    }
+
+    return retval;
+}
+
 gint
 _pygi_g_type_info_check_object (GITypeInfo *type_info,
                                 PyObject   *object,
@@ -181,8 +272,18 @@ _pygi_g_type_info_check_object (GITypeInfo *type_info,
         case GI_TYPE_TAG_BOOLEAN:
             /* No check; every Python object has a truth value. */
             break;
-        case GI_TYPE_TAG_INT8:
         case GI_TYPE_TAG_UINT8:
+            /* UINT8 types can be characters */
+            if (PYGLIB_PyBytes_Check(object)) {
+                if (PYGLIB_PyBytes_Size(object) != 1) {
+                    PyErr_Format (PyExc_TypeError, "Must be a single character");
+                    retval = 0;
+                    break;
+                }
+
+                break;
+            }
+        case GI_TYPE_TAG_INT8:
         case GI_TYPE_TAG_INT16:
         case GI_TYPE_TAG_UINT16:
         case GI_TYPE_TAG_INT32:
@@ -204,7 +305,7 @@ _pygi_g_type_info_check_object (GITypeInfo *type_info,
             if (type_tag == GI_TYPE_TAG_FLOAT || type_tag == GI_TYPE_TAG_DOUBLE) {
                 number = PyNumber_Float (object);
             } else {
-                number = PyNumber_Int (object);
+                number = PYGLIB_PyNumber_Long (object);
             }
 
             _pygi_g_type_tag_py_bounds (type_tag, &lower, &upper);
@@ -215,8 +316,8 @@ _pygi_g_type_info_check_object (GITypeInfo *type_info,
             }
 
             /* Check bounds */
-            if (PyObject_Compare (lower, number) > 0
-                    || PyObject_Compare (upper, number) < 0) {
+            if (PyObject_RichCompareBool (lower, number, Py_GT)
+                    || PyObject_RichCompareBool (upper, number, Py_LT)) {
                 PyObject *lower_str;
                 PyObject *upper_str;
 
@@ -232,10 +333,30 @@ _pygi_g_type_info_check_object (GITypeInfo *type_info,
                     goto check_number_error_release;
                 }
 
+#if PY_VERSION_HEX < 0x03000000
                 PyErr_Format (PyExc_ValueError, "Must range from %s to %s",
                               PyString_AS_STRING (lower_str),
                               PyString_AS_STRING (upper_str));
+#else
+                {
+                    PyObject *lower_pybytes_obj = PyUnicode_AsUTF8String (lower_str);
+                    if (!lower_pybytes_obj)
+                        goto utf8_fail;
 
+                    PyObject *upper_pybytes_obj = PyUnicode_AsUTF8String (upper_str);                    
+                    if (!upper_pybytes_obj) {
+                        Py_DECREF(lower_pybytes_obj);
+                        goto utf8_fail;
+                    }
+
+                    PyErr_Format (PyExc_ValueError, "Must range from %s to %s",
+                                  PyBytes_AsString (lower_pybytes_obj),
+                                  PyBytes_AsString (upper_pybytes_obj));
+                    Py_DECREF (lower_pybytes_obj);
+                    Py_DECREF (upper_pybytes_obj);
+                }
+utf8_fail:
+#endif
                 retval = 0;
 
 check_number_error_release:
@@ -251,24 +372,43 @@ check_number_release:
         }
         case GI_TYPE_TAG_GTYPE:
         {
-            gint is_instance;
-
-            is_instance = PyObject_IsInstance (object, (PyObject *) &PyGTypeWrapper_Type);
-            if (is_instance < 0) {
-                retval = -1;
-                break;
-            }
-
-            if (!is_instance && (!PyType_Check (object) || pyg_type_from_object (object) == 0)) {
+            if (pyg_type_from_object (object) == 0) {
                 PyErr_Format (PyExc_TypeError, "Must be gobject.GType, not %s",
                               object->ob_type->tp_name);
                 retval = 0;
             }
             break;
         }
+        case GI_TYPE_TAG_UNICHAR:
+        {
+            Py_ssize_t size;
+            if (PyUnicode_Check (object)) {
+                size = PyUnicode_GET_SIZE (object);
+#if PY_VERSION_HEX < 0x03000000
+            } else if (PyString_Check (object)) {
+                PyObject *pyuni = PyUnicode_FromEncodedObject (object, "UTF-8", "strict");
+                size = PyUnicode_GET_SIZE (pyuni);
+                Py_DECREF(pyuni);
+#endif
+            } else {
+                PyErr_Format (PyExc_TypeError, "Must be string, not %s",
+                              object->ob_type->tp_name);
+                retval = 0;
+                break;
+            }
+
+            if (size != 1) {
+                PyErr_Format (PyExc_TypeError, "Must be a one character string, not %ld characters",
+                              size);
+                retval = 0;
+                break;
+            }
+
+            break;
+        }
         case GI_TYPE_TAG_UTF8:
         case GI_TYPE_TAG_FILENAME:
-            if (!PyString_Check (object)) {
+            if (!PYGLIB_PyBaseString_Check (object) ) {
                 PyErr_Format (PyExc_TypeError, "Must be string, not %s",
                               object->ob_type->tp_name);
                 retval = 0;
@@ -305,6 +445,10 @@ check_number_release:
             item_type_info = g_type_info_get_param_type (type_info, 0);
             g_assert (item_type_info != NULL);
 
+            /* FIXME: This is insain.  We really should only check the first
+             *        object and perhaps have a debugging mode.  Large arrays
+             *        will cause apps to slow to a crawl.
+             */
             for (i = 0; i < length; i++) {
                 PyObject *item;
 
@@ -334,77 +478,11 @@ check_number_release:
         case GI_TYPE_TAG_INTERFACE:
         {
             GIBaseInfo *info;
-            GIInfoType info_type;
 
             info = g_type_info_get_interface (type_info);
             g_assert (info != NULL);
 
-            info_type = g_base_info_get_type (info);
-
-            switch (info_type) {
-                case GI_INFO_TYPE_CALLBACK:
-                    if (!PyCallable_Check (object)) {
-                        PyErr_Format (PyExc_TypeError, "Must be callable, not %s",
-                                      object->ob_type->tp_name);
-                        retval = 0;
-                    }
-                    break;
-                case GI_INFO_TYPE_ENUM:
-                    retval = _pygi_g_registered_type_info_check_object (
-                                 (GIRegisteredTypeInfo *) info, TRUE, object);
-                    break;
-                case GI_INFO_TYPE_FLAGS:
-                    if (PyNumber_Check (object)) {
-                        /* Accept 0 as a valid flag value */
-                        PyObject *number = PyNumber_Int (object);
-                        if (number == NULL)
-                            PyErr_Clear();
-                        else {
-                            long value = PyInt_AsLong (number);
-                            if (value == 0)
-                                break;
-                            else if (value == -1)
-                                PyErr_Clear();
-                        }
-                    }
-                    retval = _pygi_g_registered_type_info_check_object (
-                                 (GIRegisteredTypeInfo *) info, TRUE, object);
-                    break;
-                case GI_INFO_TYPE_STRUCT:
-                {
-                    GType type;
-
-                    /* Handle special cases. */
-                    type = g_registered_type_info_get_g_type ( (GIRegisteredTypeInfo *) info);
-                    if (g_type_is_a (type, G_TYPE_VALUE)) {
-                        GType object_type;
-                        object_type = pyg_type_from_object ( (PyObject *) object->ob_type);
-                        if (object_type == G_TYPE_INVALID) {
-                            PyErr_Format (PyExc_TypeError, "Must be of a known GType, not %s",
-                                          object->ob_type->tp_name);
-                            retval = 0;
-                        }
-                        break;
-                    } else if (g_type_is_a (type, G_TYPE_CLOSURE)) {
-                        if (!PyCallable_Check (object)) {
-                            PyErr_Format (PyExc_TypeError, "Must be callable, not %s",
-                                          object->ob_type->tp_name);
-                            retval = 0;
-                        }
-                        break;
-                    }
-
-                    /* Fallback. */
-                }
-                case GI_INFO_TYPE_BOXED:
-                case GI_INFO_TYPE_INTERFACE:
-                case GI_INFO_TYPE_OBJECT:
-                case GI_INFO_TYPE_UNION:
-                    retval = _pygi_g_registered_type_info_check_object ( (GIRegisteredTypeInfo *) info, TRUE, object);
-                    break;
-                default:
-                    g_assert_not_reached();
-            }
+            retval = _pygi_g_type_interface_check_object(info, object);
 
             g_base_info_unref (info);
             break;
@@ -466,7 +544,8 @@ check_number_release:
             GITypeInfo *value_type_info;
             Py_ssize_t i;
 
-            if (!PyMapping_Check (object)) {
+            keys = PyMapping_Keys (object);
+            if (keys == NULL) {
                 PyErr_Format (PyExc_TypeError, "Must be mapping, not %s",
                               object->ob_type->tp_name);
                 retval = 0;
@@ -475,12 +554,7 @@ check_number_release:
 
             length = PyMapping_Length (object);
             if (length < 0) {
-                retval = -1;
-                break;
-            }
-
-            keys = PyMapping_Keys (object);
-            if (keys == NULL) {
+                Py_DECREF (keys);
                 retval = -1;
                 break;
             }
@@ -540,8 +614,8 @@ check_number_release:
 }
 
 GArray *
-_pygi_argument_to_array (GArgument  *arg,
-                         GArgument  *args[],
+_pygi_argument_to_array (GIArgument  *arg,
+                         GIArgument  *args[],
                          GITypeInfo *type_info,
                          gboolean is_method)
 {
@@ -572,12 +646,6 @@ _pygi_argument_to_array (GArgument  *arg,
             length_arg_pos = g_type_info_get_array_length (type_info);
             g_assert (length_arg_pos >= 0);
 
-            if (is_method) {
-                length_arg_pos--;
-            }
-
-            g_assert (length_arg_pos >= 0);
-
             /* FIXME: Take into account the type of the argument. */
             length = args[length_arg_pos]->v_int;
         }
@@ -593,15 +661,15 @@ _pygi_argument_to_array (GArgument  *arg,
     return g_array;
 }
 
-GArgument
+GIArgument
 _pygi_argument_from_object (PyObject   *object,
                             GITypeInfo *type_info,
                             GITransfer  transfer)
 {
-    GArgument arg;
+    GIArgument arg;
     GITypeTag type_tag;
 
-    memset(&arg, 0, sizeof(GArgument));
+    memset(&arg, 0, sizeof(GIArgument));
     type_tag = g_type_info_get_tag (type_info);
 
     switch (type_tag) {
@@ -614,20 +682,25 @@ _pygi_argument_from_object (PyObject   *object,
             arg.v_boolean = PyObject_IsTrue (object);
             break;
         }
-        case GI_TYPE_TAG_INT8:
         case GI_TYPE_TAG_UINT8:
+            if (PYGLIB_PyBytes_Check(object)) {
+                arg.v_long = (long)(PYGLIB_PyBytes_AsString(object)[0]);
+                break;
+            }
+
+        case GI_TYPE_TAG_INT8:
         case GI_TYPE_TAG_INT16:
         case GI_TYPE_TAG_UINT16:
         case GI_TYPE_TAG_INT32:
         {
             PyObject *int_;
 
-            int_ = PyNumber_Int (object);
+            int_ = PYGLIB_PyNumber_Long (object);
             if (int_ == NULL) {
                 break;
             }
 
-            arg.v_long = PyInt_AsLong (int_);
+            arg.v_long = PYGLIB_PyLong_AsLong (int_);
 
             Py_DECREF (int_);
 
@@ -639,16 +712,17 @@ _pygi_argument_from_object (PyObject   *object,
             PyObject *number;
             guint64 value;
 
-            number = PyNumber_Int (object);
+            number = PYGLIB_PyNumber_Long (object);
             if (number == NULL) {
                 break;
             }
 
+#if PY_VERSION_HEX < 0x03000000
             if (PyInt_Check (number)) {
                 value = PyInt_AS_LONG (number);
-            } else {
-                value = PyLong_AsUnsignedLongLong (number);
-            }
+            } else
+#endif
+            value = PyLong_AsUnsignedLongLong (number);
 
             arg.v_uint64 = value;
 
@@ -661,16 +735,17 @@ _pygi_argument_from_object (PyObject   *object,
             PyObject *number;
             gint64 value;
 
-            number = PyNumber_Int (object);
+            number = PYGLIB_PyNumber_Long (object);
             if (number == NULL) {
                 break;
             }
 
+#if PY_VERSION_HEX < 0x03000000
             if (PyInt_Check (number)) {
                 value = PyInt_AS_LONG (number);
-            } else {
-                value = PyLong_AsLongLong (number);
-            }
+            } else
+#endif
+            value = PyLong_AsLongLong (number);
 
             arg.v_int64 = value;
 
@@ -712,32 +787,101 @@ _pygi_argument_from_object (PyObject   *object,
 
             break;
         }
+        case GI_TYPE_TAG_UNICHAR:
+        {
+            gchar *string;
+
+            if (object == Py_None) {
+                arg.v_uint32 = 0;
+                break;
+            }
+
+#if PY_VERSION_HEX < 0x03000000
+            if (PyUnicode_Check(object)) {
+                 PyObject *pystr_obj = PyUnicode_AsUTF8String (object);
+
+                 if (!pystr_obj)
+                     break;
+
+                 string = g_strdup(PyString_AsString (pystr_obj));
+                 Py_DECREF(pystr_obj);
+            } else {
+                 string = g_strdup(PyString_AsString (object));
+            }
+#else
+            {
+                PyObject *pybytes_obj = PyUnicode_AsUTF8String (object);
+                if (!pybytes_obj)
+                    break;
+
+                string = g_strdup(PyBytes_AsString (pybytes_obj));
+                Py_DECREF (pybytes_obj);
+            }
+#endif
+
+            arg.v_uint32 = g_utf8_get_char (string);
+
+            break;
+        }
         case GI_TYPE_TAG_UTF8:
         {
-            const gchar *string;
+            gchar *string;
 
             if (object == Py_None) {
                 arg.v_string = NULL;
                 break;
             }
+#if PY_VERSION_HEX < 0x03000000
+            if (PyUnicode_Check(object)) {
+                 PyObject *pystr_obj = PyUnicode_AsUTF8String (object);
+                 
+                 if (!pystr_obj)
+                     break;
 
-            string = PyString_AsString (object);
+                 string = g_strdup(PyString_AsString (pystr_obj));
+                 Py_DECREF(pystr_obj);
+            } else {
+                 string = g_strdup(PyString_AsString (object));
+            }
+#else
+            {
+                PyObject *pybytes_obj = PyUnicode_AsUTF8String (object);
+                if (!pybytes_obj)
+                    break;
+
+                string = g_strdup(PyBytes_AsString (pybytes_obj));
+                Py_DECREF (pybytes_obj);
+            }
+#endif
+            arg.v_string = string;
 
-            /* Don't need to check for errors, since g_strdup is NULL-proof. */
-            arg.v_string = g_strdup (string);
             break;
         }
         case GI_TYPE_TAG_FILENAME:
         {
             GError *error = NULL;
-            const gchar *string;
+            gchar *string;
+
+#if PY_VERSION_HEX < 0x03000000
+            string = g_strdup(PyString_AsString (object));
+#else
+            {
+                PyObject *pybytes_obj = PyUnicode_AsUTF8String (object);
+                if (!pybytes_obj)
+                    break;
+
+                string = g_strdup(PyBytes_AsString (pybytes_obj));
+                Py_DECREF (pybytes_obj);
+            }
+#endif
 
-            string = PyString_AsString (object);
             if (string == NULL) {
                 break;
             }
 
             arg.v_string = g_filename_from_utf8 (string, -1, NULL, NULL, &error);
+            g_free(string);
+
             if (arg.v_string == NULL) {
                 PyErr_SetString (PyExc_Exception, error->message);
                 /* TODO: Convert the error to an exception. */
@@ -777,11 +921,20 @@ _pygi_argument_from_object (PyObject   *object,
                 break;
             }
 
+            if (g_type_info_get_tag (item_type_info) == GI_TYPE_TAG_UINT8 &&
+                PYGLIB_PyBytes_Check(object)) {
+
+                memcpy(array->data, PYGLIB_PyBytes_AsString(object), length);
+                array->len = length;
+                goto array_success;
+            }
+
+
             item_transfer = transfer == GI_TRANSFER_CONTAINER ? GI_TRANSFER_NOTHING : transfer;
 
             for (i = 0; i < length; i++) {
                 PyObject *py_item;
-                GArgument item;
+                GIArgument item;
 
                 py_item = PySequence_GetItem (object, i);
                 if (py_item == NULL) {
@@ -801,7 +954,7 @@ _pygi_argument_from_object (PyObject   *object,
 
 array_item_error:
                 /* Free everything we have converted so far. */
-                _pygi_argument_release ( (GArgument *) &array, type_info,
+                _pygi_argument_release ( (GIArgument *) &array, type_info,
                                          GI_TRANSFER_NOTHING, GI_DIRECTION_IN);
                 array = NULL;
 
@@ -809,6 +962,7 @@ array_item_error:
                 break;
             }
 
+array_success:
             arg.v_pointer = array;
 
             g_base_info_unref ( (GIBaseInfo *) item_type_info);
@@ -846,7 +1000,7 @@ array_item_error:
                         GType object_type;
                         gint retval;
 
-                        object_type = pyg_type_from_object ( (PyObject *) object->ob_type);
+                        object_type = pyg_type_from_object_strict ( (PyObject *) object->ob_type, FALSE);
                         if (object_type == G_TYPE_INVALID) {
                             PyErr_SetString (PyExc_RuntimeError, "unable to retrieve object's GType");
                             break;
@@ -900,12 +1054,12 @@ array_item_error:
                 {
                     PyObject *int_;
 
-                    int_ = PyNumber_Int (object);
+                    int_ = PYGLIB_PyNumber_Long (object);
                     if (int_ == NULL) {
                         break;
                     }
 
-                    arg.v_long = PyInt_AsLong (int_);
+                    arg.v_long = PYGLIB_PyLong_AsLong (int_);
 
                     Py_DECREF (int_);
 
@@ -956,7 +1110,7 @@ array_item_error:
 
             for (i = length - 1; i >= 0; i--) {
                 PyObject *py_item;
-                GArgument item;
+                GIArgument item;
 
                 py_item = PySequence_GetItem (object, i);
                 if (py_item == NULL) {
@@ -981,7 +1135,7 @@ array_item_error:
 
 list_item_error:
                 /* Free everything we have converted so far. */
-                _pygi_argument_release ( (GArgument *) &list, type_info,
+                _pygi_argument_release ( (GIArgument *) &list, type_info,
                                          GI_TRANSFER_NOTHING, GI_DIRECTION_IN);
                 list = NULL;
 
@@ -1061,8 +1215,8 @@ list_item_error:
             for (i = 0; i < length; i++) {
                 PyObject *py_key;
                 PyObject *py_value;
-                GArgument key;
-                GArgument value;
+                GIArgument key;
+                GIArgument value;
 
                 py_key = PyList_GET_ITEM (keys, i);
                 py_value = PyList_GET_ITEM (values, i);
@@ -1083,7 +1237,7 @@ list_item_error:
 
 hash_table_item_error:
                 /* Free everything we have converted so far. */
-                _pygi_argument_release ( (GArgument *) &hash_table, type_info,
+                _pygi_argument_release ( (GIArgument *) &hash_table, type_info,
                                          GI_TRANSFER_NOTHING, GI_DIRECTION_IN);
                 hash_table = NULL;
 
@@ -1110,7 +1264,7 @@ hash_table_release:
 }
 
 PyObject *
-_pygi_argument_to_object (GArgument  *arg,
+_pygi_argument_to_object (GIArgument  *arg,
                           GITypeInfo *type_info,
                           GITransfer transfer)
 {
@@ -1135,27 +1289,27 @@ _pygi_argument_to_object (GArgument  *arg,
         }
         case GI_TYPE_TAG_INT8:
         {
-            object = PyInt_FromLong (arg->v_int8);
+            object = PYGLIB_PyLong_FromLong (arg->v_int8);
             break;
         }
         case GI_TYPE_TAG_UINT8:
         {
-            object = PyInt_FromLong (arg->v_uint8);
+            object = PYGLIB_PyLong_FromLong (arg->v_uint8);
             break;
         }
         case GI_TYPE_TAG_INT16:
         {
-            object = PyInt_FromLong (arg->v_int16);
+            object = PYGLIB_PyLong_FromLong (arg->v_int16);
             break;
         }
         case GI_TYPE_TAG_UINT16:
         {
-            object = PyInt_FromLong (arg->v_uint16);
+            object = PYGLIB_PyLong_FromLong (arg->v_uint16);
             break;
         }
         case GI_TYPE_TAG_INT32:
         {
-            object = PyInt_FromLong (arg->v_int32);
+            object = PYGLIB_PyLong_FromLong (arg->v_int32);
             break;
         }
         case GI_TYPE_TAG_UINT32:
@@ -1188,6 +1342,27 @@ _pygi_argument_to_object (GArgument  *arg,
             object = pyg_type_wrapper_new ( (GType) arg->v_long);
             break;
         }
+        case GI_TYPE_TAG_UNICHAR:
+        {
+            /* Preserve the bidirectional mapping between 0 and "" */
+            if (arg->v_uint32 == 0) {
+                object = PYGLIB_PyUnicode_FromString ("");
+            } else if (g_unichar_validate (arg->v_uint32)) {
+                gchar utf8[6];
+                gint bytes;
+
+                bytes = g_unichar_to_utf8 (arg->v_uint32, utf8);
+                object = PYGLIB_PyUnicode_FromStringAndSize ((char*)utf8, bytes);
+            } else {
+                /* TODO: Convert the error to an exception. */
+                PyErr_Format (PyExc_TypeError,
+                              "Invalid unicode codepoint %" G_GUINT32_FORMAT,
+                              arg->v_uint32);
+                object = Py_None;
+                Py_INCREF (object);
+            }
+            break;
+        }
         case GI_TYPE_TAG_UTF8:
             if (arg->v_string == NULL) {
                 object = Py_None;
@@ -1195,7 +1370,7 @@ _pygi_argument_to_object (GArgument  *arg,
                 break;
             }
 
-            object = PyString_FromString (arg->v_string);
+            object = PYGLIB_PyUnicode_FromString (arg->v_string);
             break;
         case GI_TYPE_TAG_FILENAME:
         {
@@ -1215,7 +1390,7 @@ _pygi_argument_to_object (GArgument  *arg,
                 break;
             }
 
-            object = PyString_FromString (string);
+            object = PYGLIB_PyUnicode_FromString (string);
 
             g_free (string);
 
@@ -1229,27 +1404,40 @@ _pygi_argument_to_object (GArgument  *arg,
             GITransfer item_transfer;
             gsize i, item_size;
 
-            if (arg->v_pointer == NULL) {
-                object = PyList_New (0);
-                break;
-            }
-
             array = arg->v_pointer;
 
-            object = PyList_New (array->len);
-            if (object == NULL) {
-                break;
-            }
-
             item_type_info = g_type_info_get_param_type (type_info, 0);
             g_assert (item_type_info != NULL);
 
             item_type_tag = g_type_info_get_tag (item_type_info);
             item_transfer = transfer == GI_TRANSFER_CONTAINER ? GI_TRANSFER_NOTHING : transfer;
+
+            if (item_type_tag == GI_TYPE_TAG_UINT8) {
+                /* Return as a byte array */
+                if (arg->v_pointer == NULL) {
+                    object = PYGLIB_PyBytes_FromString ("");
+                    break;
+                }
+
+                object = PYGLIB_PyBytes_FromStringAndSize(array->data, array->len);
+                break;
+
+            } else {
+                if (arg->v_pointer == NULL) {
+                    object = PyList_New (0);
+                    break;
+                }
+
+                object = PyList_New (array->len);
+                if (object == NULL) {
+                    break;
+                }
+
+            }
             item_size = g_array_get_element_size (array);
 
             for (i = 0; i < array->len; i++) {
-                GArgument item;
+                GIArgument item;
                 PyObject *py_item;
                 gboolean is_struct = FALSE;
 
@@ -1266,9 +1454,9 @@ _pygi_argument_to_object (GArgument  *arg,
                 }
 
                 if (is_struct) {
-                    item.v_pointer = &_g_array_index (array, GArgument, i);
+                    item.v_pointer = &_g_array_index (array, GIArgument, i);
                 } else {
-                    memcpy (&item, &_g_array_index (array, GArgument, i), item_size);
+                    memcpy (&item, &_g_array_index (array, GIArgument, i), item_size);
                 }
 
                 py_item = _pygi_argument_to_object (&item, item_type_info, item_transfer);
@@ -1322,6 +1510,8 @@ _pygi_argument_to_object (GArgument  *arg,
                     type = g_registered_type_info_get_g_type ( (GIRegisteredTypeInfo *) info);
                     if (g_type_is_a (type, G_TYPE_VALUE)) {
                         object = pyg_value_as_pyobject (arg->v_pointer, FALSE);
+                    } else if (g_struct_info_is_foreign (info)) {
+                        object = pygi_struct_foreign_convert_from_g_argument (type_info, arg->v_pointer);
                     } else if (g_type_is_a (type, G_TYPE_BOXED)) {
                         PyObject *py_type;
 
@@ -1345,8 +1535,6 @@ _pygi_argument_to_object (GArgument  *arg,
                         }
 
                         Py_XDECREF (py_type);
-                    } else if ( (type == G_TYPE_NONE) && (g_struct_info_is_foreign (info))) {
-                        object = pygi_struct_foreign_convert_from_g_argument (type_info, arg->v_pointer);
                     } else if (type == G_TYPE_NONE) {
                         PyObject *py_type;
 
@@ -1355,15 +1543,12 @@ _pygi_argument_to_object (GArgument  *arg,
                             break;
                         }
 
-                        if (transfer != GI_TRANSFER_NOTHING)
-                            g_warning ("Transfer mode should be set to None for "
-                                       "struct types as there is no way to free "
-                                       "them safely.  Ignoring transfer mode "
-                                       "to prevent a potential invalid free. "
-                                       "This may cause a leak in your application.");
-
+                        /* Only structs created in invoke can be safely marked
+                         * GI_TRANSFER_EVERYTHING. Trust that invoke has
+                         * filtered correctly
+                         */
                         object = _pygi_struct_new ( (PyTypeObject *) py_type, arg->v_pointer,
-                                                    FALSE);
+                                                    transfer == GI_TRANSFER_EVERYTHING);
 
                         Py_DECREF (py_type);
                     } else {
@@ -1446,7 +1631,7 @@ _pygi_argument_to_object (GArgument  *arg,
             item_transfer = transfer == GI_TRANSFER_CONTAINER ? GI_TRANSFER_NOTHING : transfer;
 
             for (i = 0; list != NULL; list = g_slist_next (list), i++) {
-                GArgument item;
+                GIArgument item;
                 PyObject *py_item;
 
                 item.v_pointer = list->data;
@@ -1470,8 +1655,8 @@ _pygi_argument_to_object (GArgument  *arg,
             GITypeInfo *value_type_info;
             GITransfer item_transfer;
             GHashTableIter hash_table_iter;
-            GArgument key;
-            GArgument value;
+            GIArgument key;
+            GIArgument value;
 
             if (arg->v_pointer == NULL) {
                 object = Py_None;
@@ -1486,9 +1671,11 @@ _pygi_argument_to_object (GArgument  *arg,
 
             key_type_info = g_type_info_get_param_type (type_info, 0);
             g_assert (key_type_info != NULL);
+            g_assert (g_type_info_get_tag (key_type_info) != GI_TYPE_TAG_VOID);
 
             value_type_info = g_type_info_get_param_type (type_info, 1);
             g_assert (value_type_info != NULL);
+            g_assert (g_type_info_get_tag (value_type_info) != GI_TYPE_TAG_VOID);
 
             item_transfer = transfer == GI_TRANSFER_CONTAINER ? GI_TRANSFER_NOTHING : transfer;
 
@@ -1533,12 +1720,13 @@ _pygi_argument_to_object (GArgument  *arg,
 }
 
 void
-_pygi_argument_release (GArgument   *arg,
+_pygi_argument_release (GIArgument   *arg,
                         GITypeInfo  *type_info,
                         GITransfer   transfer,
                         GIDirection  direction)
 {
     GITypeTag type_tag;
+    gboolean is_out = (direction == GI_DIRECTION_OUT || direction == GI_DIRECTION_INOUT);
 
     type_tag = g_type_info_get_tag (type_info);
 
@@ -1558,6 +1746,7 @@ _pygi_argument_release (GArgument   *arg,
         case GI_TYPE_TAG_FLOAT:
         case GI_TYPE_TAG_DOUBLE:
         case GI_TYPE_TAG_GTYPE:
+        case GI_TYPE_TAG_UNICHAR:
             break;
         case GI_TYPE_TAG_FILENAME:
         case GI_TYPE_TAG_UTF8:
@@ -1590,8 +1779,8 @@ _pygi_argument_release (GArgument   *arg,
 
                 /* Free the items */
                 for (i = 0; i < array->len; i++) {
-                    GArgument *item;
-                    item = &_g_array_index (array, GArgument, i);
+                    GIArgument *item;
+                    item = &_g_array_index (array, GIArgument, i);
                     _pygi_argument_release (item, item_type_info, item_transfer, direction);
                 }
 
@@ -1649,7 +1838,7 @@ _pygi_argument_release (GArgument   *arg,
                         }
                     } else if (g_struct_info_is_foreign ( (GIStructInfo*) info)) {
                         if (direction == GI_DIRECTION_OUT && transfer == GI_TRANSFER_EVERYTHING) {
-                            pygi_struct_foreign_release_g_argument (transfer, type_info, arg);
+                            pygi_struct_foreign_release (info, arg->v_pointer);
                         }
                     } else if (g_type_is_a (type, G_TYPE_BOXED)) {
                     } else if (g_type_is_a (type, G_TYPE_POINTER) || type == G_TYPE_NONE) {
@@ -1666,7 +1855,7 @@ _pygi_argument_release (GArgument   *arg,
                     if (arg->v_pointer == NULL) {
                         return;
                     }
-                    if (direction == GI_DIRECTION_OUT && transfer == GI_TRANSFER_EVERYTHING) {
+                    if (is_out && transfer == GI_TRANSFER_EVERYTHING) {
                         g_object_unref (arg->v_pointer);
                     }
                     break;
@@ -1701,7 +1890,7 @@ _pygi_argument_release (GArgument   *arg,
 
                 /* Free the items */
                 for (item = list; item != NULL; item = g_slist_next (item)) {
-                    _pygi_argument_release ( (GArgument *) &item->data, item_type_info,
+                    _pygi_argument_release ( (GIArgument *) &item->data, item_type_info,
                                              item_transfer, direction);
                 }
 
@@ -1754,9 +1943,9 @@ _pygi_argument_release (GArgument   *arg,
 
                 g_hash_table_iter_init (&hash_table_iter, hash_table);
                 while (g_hash_table_iter_next (&hash_table_iter, &key, &value)) {
-                    _pygi_argument_release ( (GArgument *) &key, key_type_info,
+                    _pygi_argument_release ( (GIArgument *) &key, key_type_info,
                                              item_transfer, direction);
-                    _pygi_argument_release ( (GArgument *) &value, value_type_info,
+                    _pygi_argument_release ( (GIArgument *) &value, value_type_info,
                                              item_transfer, direction);
                 }
 
